@@ -3,7 +3,7 @@
  *
  * Copyright (C) 2010-2011  Tetsuo Handa <penguin-kernel@I-love.SAKURA.ne.jp>
  *
- * Version: 1.0.15   2011/06/20
+ * Version: 1.0.21   2011/10/25
  */
 
 #include "internal.h"
@@ -16,8 +16,27 @@ static void ccs_task_security_gc(void);
 static int ccs_copy_cred_security(const struct cred *new,
 				  const struct cred *old, gfp_t gfp);
 static struct ccs_security *ccs_find_cred_security(const struct cred *cred);
-static void (*ccs___put_task_struct) (struct task_struct *t);
-#endif
+static DEFINE_SPINLOCK(ccs_task_security_list_lock);
+static atomic_t ccs_in_execve_tasks = ATOMIC_INIT(0);
+/*
+ * List of "struct ccs_security" for "struct pid".
+ *
+ * All instances on this list is guaranteed that "struct ccs_security"->pid !=
+ * NULL. Also, instances on this list that are in execve() are guaranteed that
+ * "struct ccs_security"->cred remembers "struct linux_binprm"->cred with a
+ * refcount on "struct linux_binprm"->cred.
+ */
+struct list_head ccs_task_security_list[CCS_MAX_TASK_SECURITY_HASH];
+/*
+ * List of "struct ccs_security" for "struct cred".
+ *
+ * Since the number of "struct cred" is nearly equals to the number of
+ * "struct pid", we allocate hash tables like ccs_task_security_list.
+ *
+ * All instances on this list are guaranteed that "struct ccs_security"->pid ==
+ * NULL and "struct ccs_security"->cred != NULL.
+ */
+static struct list_head ccs_cred_security_list[CCS_MAX_TASK_SECURITY_HASH];
 
 /* Dummy security context for avoiding NULL pointer dereference. */
 static struct ccs_security ccs_oom_security = {
@@ -28,6 +47,12 @@ static struct ccs_security ccs_oom_security = {
 static struct ccs_security ccs_default_security = {
 	.ccs_domain_info = &ccs_kernel_domain
 };
+#else
+/* Dummy security context for avoiding NULL pointer dereference. */
+extern struct ccs_security ccs_oom_security;
+/* Dummy security context for avoiding NULL pointer dereference. */
+extern struct ccs_security ccs_default_security;
+#endif
 
 /* For exporting variables and functions. */
 struct ccsecurity_exports ccsecurity_exports;
@@ -52,116 +77,6 @@ static struct security_operations original_security_ops /* = *security_ops; */;
 #define CCS_INODE_HOOK_HAS_MNT
 #endif
 
-/**
- * ccs_add_task_security - Add "struct ccs_security" to list.
- *
- * @ptr:  Pointer to "struct ccs_security".
- * @list: Pointer to "struct list_head".
- *
- * Returns nothing.
- */
-static void ccs_add_task_security(struct ccs_security *ptr,
-				  struct list_head *list)
-{
-	unsigned long flags;
-	spin_lock_irqsave(&ccs_task_security_list_lock, flags);
-	list_add_rcu(&ptr->list, list);
-	spin_unlock_irqrestore(&ccs_task_security_list_lock, flags);
-}
-
-
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 8)
-
-/**
- * ccs_rcu_free - RCU callback for releasing "struct ccs_security".
- *
- * @rcu: Pointer to "struct rcu_head".
- *
- * Returns nothing.
- */
-static void ccs_rcu_free(struct rcu_head *rcu)
-{
-	struct ccs_security *ptr = container_of(rcu, typeof(*ptr), rcu);
-	struct ccs_execve *ee = ptr->ee;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
-	/*
-	 * If this security context was used by "struct task_struct" and
-	 * remembers "struct cred" in "struct linux_binprm", it indicates that
-	 * that "struct task_struct" exited immediately after do_execve() has
-	 * failed.
-	 */
-	if (ptr->task && ptr->cred) {
-		/*
-		printk(KERN_DEBUG
-		       "Dropping refcount on \"struct cred\" in "
-		       "\"struct linux_binprm\" because some "
-		       "\"struct task_struct\" has exit()ed immediately after "
-		       "do_execve() has failed.\n");
-		*/
-		put_cred(ptr->cred);
-	}
-#endif
-	if (ee) {
-		/*
-		printk(KERN_DEBUG
-		       "Releasing memory in \"struct ccs_execve\" because "
-		       "some \"struct task_struct\" has exit()ed immediately "
-		       "after do_execve() has failed.\n");
-		*/
-		kfree(ee->handler_path);
-		kfree(ee->tmp);
-		kfree(ee->dump.data);
-		kfree(ee);
-	}
-	kfree(ptr);
-}
-
-#else
-
-/**
- * ccs_rcu_free - RCU callback for releasing "struct ccs_security".
- *
- * @arg: Pointer to "void".
- *
- * Returns nothing.
- */
-static void ccs_rcu_free(void *arg)
-{
-	struct ccs_security *ptr = arg;
-	struct ccs_execve *ee = ptr->ee;
-	if (ee) {
-		kfree(ee->handler_path);
-		kfree(ee->tmp);
-		kfree(ee->dump.data);
-		kfree(ee);
-	}
-	kfree(ptr);
-}
-
-#endif
-
-/**
- * ccs_del_security - Release "struct ccs_security".
- *
- * @ptr: Pointer to "struct ccs_security".
- *
- * Returns nothing.
- */
-static void ccs_del_security(struct ccs_security *ptr)
-{
-	unsigned long flags;
-	if (ptr == &ccs_default_security || ptr == &ccs_oom_security)
-		return;
-	spin_lock_irqsave(&ccs_task_security_list_lock, flags);
-	list_del_rcu(&ptr->list);
-	spin_unlock_irqrestore(&ccs_task_security_list_lock, flags);
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 8)
-	call_rcu(&ptr->rcu, ccs_rcu_free);
-#else
-	call_rcu(&ptr->rcu, ccs_rcu_free, ptr);
-#endif
-}
-
 /**
  * ccs_clear_execve - Release memory used by do_execve().
  *
@@ -186,20 +101,94 @@ static void ccs_clear_execve(int ret, struct ccs_security *security)
 	 */
 	put_cred(security->cred);
 	security->cred = NULL;
+	atomic_dec(&ccs_in_execve_tasks);
 #endif
-	ee->reader_idx = ccs_read_lock();
 	ccs_finish_execve(ret, ee);
 }
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
 
-/*
- * List of "struct ccs_security" for "struct cred".
+/**
+ * ccs_rcu_free - RCU callback for releasing "struct ccs_security".
  *
- * Since the number of "struct cred" is nearly equals to the number of
- * "struct task_struct", we allocate hash tables like ccs_task_security_list.
+ * @rcu: Pointer to "struct rcu_head".
+ *
+ * Returns nothing.
  */
-static struct list_head ccs_cred_security_list[CCS_MAX_TASK_SECURITY_HASH];
+static void ccs_rcu_free(struct rcu_head *rcu)
+{
+	struct ccs_security *ptr = container_of(rcu, typeof(*ptr), rcu);
+	struct ccs_execve *ee = ptr->ee;
+	/*
+	 * If this security context was associated with "struct pid" and
+	 * remembers "struct cred" in "struct linux_binprm", it indicates that
+	 * a "struct task_struct" associated with this security context exited
+	 * immediately after do_execve() has failed.
+	 */
+	if (ptr->pid && ptr->cred) {
+#ifdef CONFIG_AKARI_DEBUG
+		static bool done;
+		if (!done) {
+			printk(KERN_INFO "AKARI: Dropping refcount on "
+			       "\"struct cred\" in \"struct linux_binprm\" "
+			       "because some \"struct task_struct\" has "
+			       "exit()ed immediately after do_execve() has "
+			       "failed.\n");
+			done = true;
+		}
+#endif
+		put_cred(ptr->cred);
+		atomic_dec(&ccs_in_execve_tasks);
+	}
+	/*
+	 * If this security context was associated with "struct pid",
+	 * drop refcount obtained by get_pid() in ccs_find_task_security().
+	 */
+	if (ptr->pid) {
+#ifdef CONFIG_AKARI_DEBUG
+		static bool done;
+		if (!done) {
+			printk(KERN_INFO "AKARI: Dropping refcount on "
+			       "\"struct pid\".\n");
+			done = true;
+		}
+#endif
+		put_pid(ptr->pid);
+	}
+	if (ee) {
+#ifdef CONFIG_AKARI_DEBUG
+		static bool done;
+		if (!done) {
+			printk(KERN_INFO "AKARI: Releasing memory in "
+			       "\"struct ccs_execve\" because some "
+			       "\"struct task_struct\" has exit()ed "
+			       "immediately after do_execve() has failed.\n");
+			done = true;
+		}
+#endif
+		kfree(ee->handler_path);
+		kfree(ee);
+	}
+	kfree(ptr);
+}
+
+/**
+ * ccs_del_security - Release "struct ccs_security".
+ *
+ * @ptr: Pointer to "struct ccs_security".
+ *
+ * Returns nothing.
+ */
+static void ccs_del_security(struct ccs_security *ptr)
+{
+	unsigned long flags;
+	if (ptr == &ccs_default_security || ptr == &ccs_oom_security)
+		return;
+	spin_lock_irqsave(&ccs_task_security_list_lock, flags);
+	list_del_rcu(&ptr->list);
+	spin_unlock_irqrestore(&ccs_task_security_list_lock, flags);
+	call_rcu(&ptr->rcu, ccs_rcu_free);
+}
 
 /**
  * ccs_add_cred_security - Add "struct ccs_security" to list.
@@ -213,6 +202,12 @@ static void ccs_add_cred_security(struct ccs_security *ptr)
 	unsigned long flags;
 	struct list_head *list = &ccs_cred_security_list
 		[hash_ptr((void *) ptr->cred, CCS_TASK_SECURITY_HASH_BITS)];
+#ifdef CONFIG_AKARI_DEBUG
+	if (ptr->pid)
+		printk(KERN_INFO "AKARI: \"struct ccs_security\"->pid != NULL"
+		       "\n");
+#endif
+	ptr->pid = NULL;
 	spin_lock_irqsave(&ccs_task_security_list_lock, flags);
 	list_add_rcu(&ptr->list, list);
 	spin_unlock_irqrestore(&ccs_task_security_list_lock, flags);
@@ -354,28 +349,6 @@ static void ccs_cred_transfer(struct cred *new, const struct cred *old)
 
 #else
 
-/**
- * ccs_copy_task_security - Allocate memory for new tasks.
- *
- * @task: Pointer to "struct task_struct".
- *
- * Returns 0 on success, negative value otherwise.
- */
-static int ccs_copy_task_security(struct task_struct *task)
-{
-	struct ccs_security *old_security = ccs_current_security();
-	struct ccs_security *new_security = kzalloc(sizeof(*new_security),
-						    GFP_KERNEL);
-	struct list_head *list = &ccs_task_security_list
-		[hash_ptr((void *) task, CCS_TASK_SECURITY_HASH_BITS)];
-	if (!new_security)
-		return -ENOMEM;
-	*new_security = *old_security;
-	new_security->task = task;
-	ccs_add_task_security(new_security, list);
-	return 0;
-}
-
 /**
  * ccs_task_alloc_security - Allocate memory for new tasks.
  *
@@ -385,13 +358,13 @@ static int ccs_copy_task_security(struct task_struct *task)
  */
 static int ccs_task_alloc_security(struct task_struct *p)
 {
-	int rc = ccs_copy_task_security(p);
+	int rc = ccs_alloc_task_security(p);
 	if (rc)
 		return rc;
 	while (!original_security_ops.task_alloc_security);
 	rc = original_security_ops.task_alloc_security(p);
 	if (rc)
-		ccs_del_security(ccs_find_task_security(p));
+		ccs_free_task_security(p);
 	return rc;
 }
 
@@ -406,7 +379,7 @@ static void ccs_task_free_security(struct task_struct *p)
 {
 	while (!original_security_ops.task_free_security);
 	original_security_ops.task_free_security(p);
-	ccs_del_security(ccs_find_task_security(p));
+	ccs_free_task_security(p);
 }
 
 /**
@@ -506,20 +479,22 @@ static int ccs_bprm_check_security(struct linux_binprm *bprm)
 		return -ENOMEM;
 	if (!security->ee) {
 		int rc;
+#ifndef CONFIG_CCSECURITY_OMIT_USERSPACE_LOADER
 		if (!ccs_policy_loaded)
 			ccs_load_policy(bprm->filename);
+#endif
 		rc = ccs_start_execve(bprm, &security->ee);
-		if (security->ee) {
-			ccs_read_unlock(security->ee->reader_idx);
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
+		if (security->ee) {
 			/*
 			 * Get refcount on "struct cred" in
 			 * "struct linux_binprm" and remember it.
 			 */
 			get_cred(bprm->cred);
 			security->cred = bprm->cred;
-#endif
+			atomic_inc(&ccs_in_execve_tasks);
 		}
+#endif
 		if (rc)
 			return rc;
 	}
@@ -1876,6 +1851,8 @@ static int ccs_file_ioctl(struct file *filp, unsigned int cmd,
 }
 
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 21) && LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 33) && defined(CONFIG_SYSCTL_SYSCALL)
+int ccs_path_permission(struct ccs_request_info *r, u8 operation,
+			const struct ccs_path_info *filename);
 
 /**
  * ccs_prepend - Copy of prepend() in fs/dcache.c.
@@ -2094,6 +2071,38 @@ out:
 
 #endif
 
+#if defined(CONFIG_ARM) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
+static int lsm_addr_calculator(struct file *file);
+static void * __init ccs_find_security_ops_on_arm(unsigned int *base);
+#endif
+
+#if defined(CONFIG_ARM) && LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 36)
+/**
+ * ccs_find_vfsmount_lock_on_arm - Find vfsmount_lock spinlock on ARM.
+ *
+ * @ip:   Address of dummy function's entry point.
+ * @addr: Address of the variable which is used within @function.
+ * @base: Address of function's entry point.
+ *
+ * Returns address of vfsmount_lock on success, NULL otherwise.
+ */
+static void * __init ccs_find_vfsmount_lock_on_arm(unsigned int *ip,
+						   unsigned long addr,
+						   unsigned int *base)
+{
+	int i;
+	for (i = 0; i < 32; ip++, i++) {
+		static unsigned int *ip4ret;
+		if (*(ip + 2 + ((*ip & 0xFFF) >> 2)) != addr)
+			continue;
+		ip = base + i;
+		ip4ret = (unsigned int *) (*(ip + 2 + ((*ip & 0xFFF) >> 2)));
+		return &ip4ret;
+	}
+	return NULL;
+}
+#endif
+
 /**
  * ccs_find_variable - Find variable's address using dummy.
  *
@@ -2122,6 +2131,14 @@ static void * __init ccs_find_variable(void *function, unsigned long addr,
 		base = __symbol_get(symbol);
 	if (!base)
 		return NULL;
+#if defined(CONFIG_ARM) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
+	if (function == lsm_addr_calculator)
+		return ccs_find_security_ops_on_arm((unsigned int *) base);
+#endif
+#if defined(CONFIG_ARM) && LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 36)
+	return ccs_find_vfsmount_lock_on_arm(function, addr,
+					     (unsigned int *) base);
+#endif
 	/* First, assume absolute adressing mode is used. */
 	for (i = 0; i < 128; i++) {
 		if (*(unsigned long *) cp == addr)
@@ -2164,7 +2181,7 @@ static struct security_operations *ccs_security_ops;
  *
  * @file: Pointer to "struct file".
  *
- * Returns return value fromfrom security_file_alloc().
+ * Returns return value from security_file_alloc().
  *
  * Never mark this function as __init in order to make sure that compiler
  * generates identical code for security_file_alloc() and this function.
@@ -2174,6 +2191,50 @@ static int lsm_addr_calculator(struct file *file)
 	return ccs_security_ops->file_alloc_security(file);
 }
 
+#ifdef CONFIG_ARM
+/**
+ * ccs_find_security_ops_on_arm - Find security_ops on ARM.
+ *
+ * @base: Address of security_file_alloc().
+ *
+ * Returns address of security_ops on success, NULL otherwise.
+ */
+static void * __init ccs_find_security_ops_on_arm(unsigned int *base)
+{
+	static unsigned int *ip4ret;
+	int i;
+	const unsigned long addr = (unsigned long) &ccs_security_ops;
+	unsigned int *ip = (unsigned int *) lsm_addr_calculator;
+	for (i = 0; i < 32; ip++, i++) {
+		if (*(ip + 2 + ((*ip & 0xFFF) >> 2)) != addr)
+			continue;
+		ip = base + i;
+		ip4ret = (unsigned int *) (*(ip + 2 + ((*ip & 0xFFF) >> 2)));
+		return &ip4ret;
+	}
+	ip = (unsigned int *) lsm_addr_calculator;
+	for (i = 0; i < 32; ip++, i++) {
+		/*
+		 * Find
+		 *   ldr r3, [pc, #offset1]
+		 *   ldr r3, [r3, #offset2]
+		 * sequence.
+		 */
+		if ((*ip & 0xFFFFF000) != 0xE59F3000 ||
+		    (*(ip + 1) & 0xFFFFF000) != 0xE5933000)
+			continue;
+		ip4ret = (unsigned int *) (*(ip + 2 + ((*ip & 0xFFF) >> 2)));
+		ip4ret += (*(ip + 1) & 0xFFF) >> 2;
+		if ((unsigned long) ip4ret != addr)
+			continue;
+		ip = base + i;
+		ip4ret = (unsigned int *) (*(ip + 2 + ((*ip & 0xFFF) >> 2)));
+		ip4ret += (*(ip + 1) & 0xFFF) >> 2;
+		return &ip4ret;
+	}
+	return NULL;
+}
+#endif
 #endif
 
 /**
@@ -2204,7 +2265,7 @@ static struct security_operations * __init ccs_find_security_ops(void)
 		printk(KERN_ERR "Can't resolve security_ops structure.\n");
 		goto out;
 	}
-	printk(KERN_INFO "&security_ops=%p\n", ptr);
+	printk(KERN_INFO "security_ops=%p\n", ptr);
 	ops = *ptr;
 	if (!ops) {
 		printk(KERN_ERR "No security_operations registered.\n");
@@ -2254,65 +2315,64 @@ out:
 #endif
 }
 
-/**
- * ccs_find___put_task_struct - Find address of __put_task_struct().
- *
- * Returns true on success, false otherwise.
- */
-static bool __init ccs_find___put_task_struct(void)
-{
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
-	void *ptr;
-	ptr = ccs_find_symbol(" __put_task_struct\n");
-	if (!ptr) {
-		printk(KERN_ERR "Can't resolve __put_task_struct().\n");
-		goto out;
-	}
-	ccs___put_task_struct = ptr;
-	printk(KERN_INFO "__put_task_struct=%p\n", ptr);
-	return true;
-out:
-	return false;
-#else
-	return true;
-#endif
-}
+#if defined(CONFIG_SMP) || defined(CONFIG_DEBUG_SPINLOCK) || LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 36)
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 3)
 
 /* Never mark this variable as __initdata . */
-static spinlock_t ccs_vfsmount_lock;
+static spinlock_t ccs_vfsmount_lock __cacheline_aligned_in_smp =
+SPIN_LOCK_UNLOCKED;
+
+static struct list_head *mount_hashtable;
+static int hash_mask, hash_bits;
 
 /**
- * lsm_umnt_tr - Dummy function which prototype is identical to umount_tree() in fs/namespace.c.
+ * hash - Copy of hash() in fs/namespace.c.
  *
  * @mnt: Pointer to "struct vfsmount".
+ * @dentry: Pointer to "struct dentry".
  *
- * Returns nothing.
+ * Returns hash value.
  */
-static void lsm_umnt_tr(struct vfsmount *mnt)
+static inline unsigned long hash(struct vfsmount *mnt, struct dentry *dentry)
 {
-	/* Nothing to do because this function is not called. */
+	unsigned long tmp = ((unsigned long) mnt / L1_CACHE_BYTES);
+	tmp += ((unsigned long) dentry / L1_CACHE_BYTES);
+	tmp = tmp + (tmp >> hash_bits);
+	return tmp & hash_mask;
 }
 
 /**
- * lsm__put_nmspce - Dummy function which does identical to __put_namespace() in fs/namespace.c.
+ * lsm_lu_mnt - Dummy function which does identical to lookup_mnt() in fs/namespace.c.
  *
- * @namespace: Pointer to "struct namespace".
+ * @mnt:    Pointer to "struct vfsmount".
+ * @dentry: Pointer to "struct dentry".
  *
- * Returns nothing.
+ * Returns pointer to "struct vfsmount".
  *
  * Never mark this function as __init in order to make sure that compiler
- * generates identical code for __put_namespace() and this function.
+ * generates identical code for lookup_mnt() and this function.
  */
-static void lsm__put_nmspce(struct namespace *namespace)
+static struct vfsmount *lsm_lu_mnt(struct vfsmount *mnt, struct dentry *dentry)
 {
-	down_write(&namespace->sem);
+	struct list_head *head = mount_hashtable + hash(mnt, dentry);
+	struct list_head *tmp = head;
+	struct vfsmount *p, *found = NULL;
+
 	spin_lock(&ccs_vfsmount_lock);
-	lsm_umnt_tr(namespace->root);
+	for (;;) {
+		tmp = tmp->next;
+		p = NULL;
+		if (tmp == head)
+			break;
+		p = list_entry(tmp, struct vfsmount, mnt_hash);
+		if (p->mnt_parent == mnt && p->mnt_mountpoint == dentry) {
+			found = mntget(p);
+			break;
+		}
+	}
 	spin_unlock(&ccs_vfsmount_lock);
-	up_write(&namespace->sem);
-	kfree(namespace);
+	return found;
 }
 
 #elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 15)
@@ -2370,7 +2430,7 @@ static void lsm_pin(struct vfsmount *mnt)
 {
 	spin_lock(&ccs_vfsmount_lock);
 	mnt->mnt_pinned++;
-	spin_lock(&ccs_vfsmount_lock);
+	spin_unlock(&ccs_vfsmount_lock);
 }
 
 #endif
@@ -2386,11 +2446,10 @@ static bool __init ccs_find_vfsmount_lock(void)
 	void *cp;
 	spinlock_t *ptr;
 	/* Guess "spinlock_t vfsmount_lock;". */
-	cp = ccs_find_variable(lsm__put_nmspce,
-			       (unsigned long) &ccs_vfsmount_lock,
-			       " __put_namespace\n");
+	cp = ccs_find_variable(lsm_lu_mnt, (unsigned long) &ccs_vfsmount_lock,
+			       " lookup_mnt\n");
 	if (!cp) {
-		printk(KERN_ERR "Can't resolve __put_namespace().\n");
+		printk(KERN_ERR "Can't resolve lookup_mnt().\n");
 		goto out;
 	}
 	/* This should be "spinlock_t *vfsmount_lock;". */
@@ -2403,13 +2462,6 @@ static bool __init ccs_find_vfsmount_lock(void)
 	printk(KERN_INFO "vfsmount_lock=%p\n", ptr);
 	return true;
 out:
-	/*
-	 * Dummy call for preventing lsm_umnt_tr() from being inlined into
-	 * lsm__put_nmspce().
-	 */
-	cp = ccs_find_variable(lsm_umnt_tr,
-			       (unsigned long) &ccs_vfsmount_lock,
-			       " __put_namespace\n");
 	return false;
 #elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 15)
 	void *cp;
@@ -2465,6 +2517,24 @@ out:
 #endif
 }
 
+#else
+
+/* Never mark this variable as __initdata . */
+static spinlock_t ccs_vfsmount_lock;
+
+/**
+ * ccs_find_vfsmount_lock - Find address of "spinlock_t vfsmount_lock".
+ *
+ * Returns true on success, false otherwise.
+ */
+static bool __init ccs_find_vfsmount_lock(void)
+{
+	ccsecurity_exports.vfsmount_lock = &ccs_vfsmount_lock;
+	return true;
+}
+
+#endif
+
 /*
  * Why not to copy all operations by "original_security_ops = *ops" ?
  * Because copying byte array is not atomic. Reader checks
@@ -2573,18 +2643,20 @@ static int __init ccs_init(void)
 {
 	struct security_operations *ops = ccs_find_security_ops();
 	if (!ops || !ccs_find_find_task_by_pid() ||
-	    !ccs_find_vfsmount_lock() || !ccs_find___put_task_struct())
+	    !ccs_find_vfsmount_lock())
 		return -EINVAL;
 #if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
 	{
 		int idx;
-		for (idx = 0; idx < CCS_MAX_TASK_SECURITY_HASH; idx++)
+		for (idx = 0; idx < CCS_MAX_TASK_SECURITY_HASH; idx++) {
 			INIT_LIST_HEAD(&ccs_cred_security_list[idx]);
+			INIT_LIST_HEAD(&ccs_task_security_list[idx]);
+		}
 	}
 #endif
 	ccs_main_init();
 	ccs_update_security_ops(ops);
-	printk(KERN_INFO "AKARI: 1.0.15   2011/06/20\n");
+	printk(KERN_INFO "AKARI: 1.0.21   2011/10/25\n");
 	printk(KERN_INFO
 	       "Access Keeping And Regulating Instrument registered.\n");
 	return 0;
@@ -2621,6 +2693,25 @@ bool ccs_used_by_cred(const struct ccs_domain_info *domain)
 	return false;
 }
 
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
+
+/**
+ * ccs_add_task_security - Add "struct ccs_security" to list.
+ *
+ * @ptr:  Pointer to "struct ccs_security".
+ * @list: Pointer to "struct list_head".
+ *
+ * Returns nothing.
+ */
+static void ccs_add_task_security(struct ccs_security *ptr,
+				  struct list_head *list)
+{
+	unsigned long flags;
+	spin_lock_irqsave(&ccs_task_security_list_lock, flags);
+	list_add_rcu(&ptr->list, list);
+	spin_unlock_irqrestore(&ccs_task_security_list_lock, flags);
+}
+
 /**
  * ccs_find_task_security - Find "struct ccs_security" for given task.
  *
@@ -2643,10 +2734,9 @@ struct ccs_security *ccs_find_task_security(const struct task_struct *task)
 	while (!list->next);
 	rcu_read_lock();
 	list_for_each_entry_rcu(ptr, list, list) {
-		if (ptr->task != task)
+		if (ptr->pid != task->pids[PIDTYPE_PID].pid)
 			continue;
 		rcu_read_unlock();
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
 		/*
 		 * Current thread needs to transit from old domain to new
 		 * domain before do_execve() succeeds in order to check
@@ -2679,19 +2769,21 @@ struct ccs_security *ccs_find_task_security(const struct task_struct *task)
 		 */
 		if (task == current && ptr->cred &&
 		    atomic_read(&ptr->cred->usage) == 1) {
-			/*
-			printk(KERN_DEBUG
-			       "pid=%u: Reverting domain transition because "
-			       "do_execve() has failed.\n", task->pid);
-			*/
+#ifdef CONFIG_AKARI_DEBUG
+			static bool done;
+			if (!done) {
+				printk(KERN_INFO "AKARI: Reverting domain "
+				       "transition because do_execve() has "
+				       "failed.\n");
+				done = true;
+			}
+#endif
 			ccs_clear_execve(-1, ptr);
 		}
-#endif
 		return ptr;
 	}
 	rcu_read_unlock();
 	if (task != current) {
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
 		/*
 		 * If a thread does nothing after fork(), caller will reach
 		 * here because "struct ccs_security" for that thread is not
@@ -2711,9 +2803,6 @@ struct ccs_security *ccs_find_task_security(const struct task_struct *task)
 		 * like get_robust_list() does.
 		 */
 		return ccs_find_cred_security(__task_cred(task));
-#else
-		return &ccs_default_security;
-#endif
 	}
 	/* Use GFP_ATOMIC because caller may have called rcu_read_lock(). */
 	ptr = kzalloc(sizeof(*ptr), GFP_ATOMIC);
@@ -2723,23 +2812,15 @@ struct ccs_security *ccs_find_task_security(const struct task_struct *task)
 		send_sig(SIGKILL, current, 0);
 		return &ccs_oom_security;
 	}
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
 	*ptr = *ccs_find_cred_security(task->cred);
-#else
-	*ptr = ccs_default_security;
-#endif
-	get_task_struct((struct task_struct *) task);
-	ptr->task = (struct task_struct *) task;
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
-	/* ptr->cred may point to garbage. I need to explicitly clear. */
+	/* We can shortcut because task == current. */
+	ptr->pid = get_pid(((struct task_struct *) task)->
+			   pids[PIDTYPE_PID].pid);
 	ptr->cred = NULL;
-#endif
 	ccs_add_task_security(ptr, list);
 	return ptr;
 }
 
-#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 29)
-
 /**
  * ccs_copy_cred_security - Allocate memory for new credentials.
  *
@@ -2794,44 +2875,39 @@ static struct ccs_security *ccs_find_cred_security(const struct cred *cred)
  *
  * Since security_task_free_security() is missing, I can't release memory
  * associated with "struct task_struct" when a task dies. Therefore, I hold
- * a reference on "struct task_struct" and runs garbage collection when I
- * became the last user who refers that "struct task_struct".
+ * a reference on "struct pid" and runs garbage collection when associated
+ * "struct task_struct" has gone.
  */
 static void ccs_task_security_gc(void)
 {
 	static DEFINE_MUTEX(lock);
-	static unsigned int counter;
+	static atomic_t gc_counter = ATOMIC_INIT(0);
 	unsigned int idx;
-	if (!mutex_trylock(&lock))
+	/*
+	 * If some process is doing execve(), try to garbage collection now.
+	 * We should kfree() memory associated with "struct ccs_security"->ee
+	 * as soon as execve() has completed in order to compensate for lack of
+	 * security_bprm_free() and security_task_free() hooks.
+	 *
+	 * Otherwise, reduce frequency for performance reason.
+	 */
+	if (!atomic_read(&ccs_in_execve_tasks) &&
+	    atomic_inc_return(&gc_counter) < 1024)
+		return;
+	atomic_set(&gc_counter, 0);
+	if (mutex_lock_interruptible(&lock))
 		return;
-	/* Checking every time is too wasteful. */
-	if (counter++ < 1024)
-		goto skip;
-	counter = 0;
 	rcu_read_lock();
 	for (idx = 0; idx < CCS_MAX_TASK_SECURITY_HASH; idx++) {
 		struct ccs_security *ptr;
 		struct list_head *list = &ccs_task_security_list[idx];
 		list_for_each_entry_rcu(ptr, list, list) {
-			struct task_struct *task =
-				(struct task_struct *) ptr->task;
-			/* Am I the last one using this task? */
-			if (atomic_read(&task->usage) != 1)
+			if (pid_task(ptr->pid, PIDTYPE_PID))
 				continue;
-			/*
-			 * We need to call put_task_struct(task); here.
-			 * However, since put_task_struct() is an inlined
-			 * function which calls __put_task_struct() and
-			 * __put_task_struct() is not exported,
-			 * we embed put_task_struct() into here.
-			 */
-			atomic_dec(&task->usage);
-			ccs___put_task_struct(task);
 			ccs_del_security(ptr);
 		}
 	}
 	rcu_read_unlock();
-skip:
 	mutex_unlock(&lock);
 }
 
